use std::collections::HashMap;
use std::ops::Neg;

use bitfinex::ticker::TradingPairTicker;
use futures_util::stream::FuturesUnordered;
use futures_util::StreamExt;
use log::{debug, error, info};
use tokio::signal::unix::Signal;
use tokio::sync::mpsc::channel;
use tokio::sync::mpsc::{Receiver, Sender};
use tokio::sync::oneshot;

use crate::connectors::{Client, ExchangeKind};
use crate::currency::SymbolPair;
use crate::events::{ActorMessage, Event, Message};
use crate::models::{ActiveOrder, OrderBook, OrderForm, OrderKind, Position, PriceTicker};
use crate::strategy::{FastOrderStrategy, OrderStrategy, PositionStrategy, TrailingStop};
use crate::BoxError;

pub type OptionUpdate = (Option<Vec<Event>>, Option<Vec<Message>>);

pub struct EventManager {
    events: Vec<Event>,
}

/******************
* PRICES
******************/

#[derive(Debug)]
pub struct PriceManager {
    receiver: Receiver<ActorMessage>,
    pair: SymbolPair,
    prices: Vec<PriceEntry>,
    client: Client,
}

pub struct PriceManagerHandle {
    sender: Sender<ActorMessage>,
}

impl PriceManagerHandle {
    async fn run_price_manager(mut manager: PriceManager) {
        while let Some(msg) = manager.receiver.recv().await {
            manager.handle_message(msg).await.unwrap();
        }
    }

    pub fn new(pair: SymbolPair, client: Client) -> Self {
        let (sender, receiver) = channel(8);

        let price_manager = PriceManager::new(receiver, pair, client);
        tokio::spawn(PriceManagerHandle::run_price_manager(price_manager));

        Self { sender }
    }

    pub async fn update(&mut self, tick: u64) -> Result<OptionUpdate, BoxError> {
        let (send, recv) = oneshot::channel();

        self.sender
            .send(ActorMessage {
                message: Message::Update { tick },
                respond_to: send,
            })
            .await?;

        Ok(recv.await?)
    }
}

impl PriceManager {
    pub fn new(receiver: Receiver<ActorMessage>, pair: SymbolPair, client: Client) -> Self {
        PriceManager {
            receiver,
            pair,
            prices: Vec::new(),
            client,
        }
    }

    pub async fn handle_message(&mut self, message: ActorMessage) -> Result<(), BoxError> {
        match message.message {
            Message::Update { tick } => {
                let a = self.update(tick).await?;
                self.add_entry(a);
            }
            _ => {}
        }

        message.respond_to.send((None, None));

        Ok(())
    }

    pub fn add_entry(&mut self, entry: PriceEntry) {
        self.prices.push(entry);
    }

    pub async fn update(&mut self, tick: u64) -> Result<PriceEntry, BoxError> {
        let current_prices = self.client.current_prices(&self.pair).await?.into();

        Ok(PriceEntry::new(
            tick,
            current_prices,
            self.pair.clone(),
            None,
            // None,
        ))
    }

    // fn add_event(&mut self, event: Event) {
    //     self.events.push(event);
    //
    //     self.dispatcher.call_event_handlers(&event, &self);
    // }
    //
    // fn add_signal(&mut self, signal: SignalKind) {
    //     self.signals.insert(self.current_tick(), signal);
    // }

    pub fn pair(&self) -> &SymbolPair {
        &self.pair
    }
}

#[derive(Clone, Debug)]
pub struct PriceEntry {
    tick: u64,
    pair: SymbolPair,
    price: PriceTicker,
    events: Option<Vec<Event>>,
    // signals: Option<Vec<SignalKind>>,
}

impl PriceEntry {
    pub fn new(
        tick: u64,
        price: PriceTicker,
        pair: SymbolPair,
        events: Option<Vec<Event>>,
        // signals: Option<Vec<SignalKind>>,
    ) -> Self {
        PriceEntry {
            tick,
            pair,
            price,
            events,
            // signals,
        }
    }

    pub fn tick(&self) -> u64 {
        self.tick
    }
    pub fn pair(&self) -> &SymbolPair {
        &self.pair
    }
    pub fn price(&self) -> PriceTicker {
        self.price
    }
    pub fn events(&self) -> &Option<Vec<Event>> {
        &self.events
    }
    // pub fn signals(&self) -> &Option<Vec<SignalKind>> {
    //     &self.signals
    // }
}

/******************
* POSITIONS
******************/

pub struct PositionManagerHandle {
    sender: Sender<ActorMessage>,
}

impl PositionManagerHandle {
    async fn run_position_manager(mut manager: PositionManager) {
        while let Some(msg) = manager.receiver.recv().await {
            manager.handle_message(msg).await.unwrap();
        }
    }

    pub fn new(pair: SymbolPair, client: Client, strategy: Box<dyn PositionStrategy>) -> Self {
        let (sender, receiver) = channel(8);

        let manager = PositionManager::new(receiver, pair, client, strategy);

        tokio::spawn(PositionManagerHandle::run_position_manager(manager));

        Self { sender }
    }

    pub async fn update(&mut self, tick: u64) -> Result<OptionUpdate, BoxError> {
        let (send, recv) = oneshot::channel();

        self.sender
            .send(ActorMessage {
                message: Message::Update { tick },
                respond_to: send,
            })
            .await?;

        let response = recv.await?;

        Ok(response)
    }
}

#[derive(Debug)]
pub struct PositionManager {
    receiver: Receiver<ActorMessage>,
    current_tick: u64,
    pair: SymbolPair,
    positions_history: HashMap<u64, Position>,
    active_position: Option<Position>,
    client: Client,
    strategy: Box<dyn PositionStrategy>,
}

impl PositionManager {
    pub fn new(
        receiver: Receiver<ActorMessage>,
        pair: SymbolPair,
        client: Client,
        strategy: Box<dyn PositionStrategy>,
    ) -> Self {
        PositionManager {
            receiver,
            current_tick: 0,
            pair,
            positions_history: HashMap::new(),
            active_position: None,
            client,
            strategy,
        }
    }

    pub fn current_tick(&self) -> u64 {
        self.current_tick
    }

    pub async fn handle_message(&mut self, msg: ActorMessage) -> Result<(), BoxError> {
        match msg.message {
            Message::Update { tick } => {
                let result = self.update(tick).await?;

                msg.respond_to.send(result);
            }
            _ => {}
        };

        Ok(())
    }

    pub async fn update(&mut self, tick: u64) -> Result<OptionUpdate, BoxError> {
        // debug!("Updating {}", self.pair);

        let opt_active_positions = self.client.active_positions(&self.pair).await?;
        self.current_tick = tick;

        // we assume there is only ONE active position per pair
        match opt_active_positions {
            // no open positions, no events and no messages returned
            None => return Ok((None, None)),

            Some(positions) => {
                // checking if there are positions open for our pair
                match positions
                    .into_iter()
                    .filter(|x| x.pair() == &self.pair)
                    .next()
                {
                    // no open positions for our pair, setting active position to none
                    None => {
                        self.active_position = None;
                        return Ok((None, None));
                    }

                    // applying strategy to open position and saving into struct
                    Some(position) => {
                        let (position_after_strategy, strategy_events, strategy_messages) =
                            self.strategy.on_new_tick(position, &self);

                        self.positions_history
                            .insert(self.current_tick(), position_after_strategy.clone());
                        self.active_position = Some(position_after_strategy);

                        return Ok((strategy_events, strategy_messages));
                    }
                }
            }
        };
    }

    pub fn position_previous_tick(&self, id: u64, tick: Option<u64>) -> Option<&Position> {
        let tick = match tick {
            Some(tick) => {
                if tick < 1 {
                    1
                } else {
                    tick
                }
            }
            None => self.current_tick() - 1,
        };

        self.positions_history
            .get(&tick)
            .filter(|x| x.id() == id)
            .and_then(|x| Some(x))
    }
}

/******************
* ORDERS
******************/

pub type TrackedPositionsMap = HashMap<u64, ActiveOrder>;

pub struct OrderManagerHandle {
    sender: Sender<ActorMessage>,
}

impl OrderManagerHandle {
    async fn run_order_manager(mut manager: OrderManager) {
        while let Some(msg) = manager.receiver.recv().await {
            manager.handle_message(msg).await.unwrap();
        }
    }

    pub fn new(pair: SymbolPair, client: Client, strategy: Box<dyn OrderStrategy>) -> Self {
        let (sender, receiver) = channel(8);

        let manager = OrderManager::new(receiver, pair, client, strategy);

        tokio::spawn(OrderManagerHandle::run_order_manager(manager));

        Self { sender }
    }

    pub async fn update(&mut self, tick: u64) -> Result<OptionUpdate, BoxError> {
        let (send, recv) = oneshot::channel();

        self.sender
            .send(ActorMessage {
                message: Message::Update { tick },
                respond_to: send,
            })
            .await?;

        Ok(recv.await?)
    }

    pub async fn close_position(
        &mut self,
        position: Position,
        order_kind: OrderKind,
    ) -> Result<OptionUpdate, BoxError> {
        let (send, recv) = oneshot::channel();

        self.sender
            .send(ActorMessage {
                message: Message::ClosePosition {
                    position,
                    order_kind,
                },
                respond_to: send,
            })
            .await?;

        Ok(recv.await?)
    }
}

pub struct OrderManager {
    receiver: Receiver<ActorMessage>,
    tracked_positions: TrackedPositionsMap,
    pair: SymbolPair,
    open_orders: Vec<ActiveOrder>,
    client: Client,
    strategy: Box<dyn OrderStrategy>,
}

impl OrderManager {
    pub fn new(
        receiver: Receiver<ActorMessage>,
        pair: SymbolPair,
        client: Client,
        strategy: Box<dyn OrderStrategy>,
    ) -> Self {
        OrderManager {
            receiver,
            pair,
            open_orders: Vec::new(),
            client,
            strategy,
            tracked_positions: HashMap::new(),
        }
    }

    pub async fn handle_message(&mut self, msg: ActorMessage) -> Result<(), BoxError> {
        match msg.message {
            Message::Update { .. } => {
                self.update();
            }
            Message::ClosePosition {
                position,
                order_kind,
            } => self.close_position(&position, &order_kind).await?,
            _ => {}
        };

        msg.respond_to.send((None, None));

        Ok(())
    }

    pub async fn close_position(
        &mut self,
        position: &Position,
        order_kind: &OrderKind,
    ) -> Result<(), BoxError> {
        let open_order = self.tracked_positions.get(&position.id());

        debug!("Closing position #{}", position.id());

        debug!("Getting current prices...");
        let order_book = self.client.order_book(&self.pair).await?;

        // checking if the position has an open order.
        // If so, the strategy method is called, otherwise we open
        // an undercut limit order at the best current price.
        match open_order {
            Some(open_order) => {
                debug!("There is an open order. Calling strategy.");
                let (_, messages) =
                    self.strategy
                        .on_position_close(open_order, position, &order_book)?;

                if let Some(messages) = messages {
                    for m in messages {
                        match m {
                            Message::ClosePosition {
                                position,
                                order_kind,
                            } => {
                                // TODO FIXME
                                match order_kind {
                                    OrderKind::Market => {
                                        info!("Closing open order with a market order.");
                                        if let Ok(_) = self
                                            .submit_market_order(1.0, position.amount().neg())
                                            .await
                                        {
                                            // cancel active order
                                            info!("Cancelling open order #{}", open_order.id);
                                            self.client.cancel_order(open_order).await?;
                                        }
                                    }
                                    _ => {}
                                }
                            }
                            _ => {
                                debug!("Received unsupported message from order strategy. Unimplemented.")
                            }
                        }
                    }
                }
            }
            None => {
                let closing_price = self.best_closing_price(&position, &order_book);

                let active_order = match order_kind {
                    OrderKind::Limit => {
                        self.submit_limit_order(closing_price, position.amount().neg())
                            .await?
                    }
                    OrderKind::ExchangeLimit => {
                        self.submit_exchange_limit_order(closing_price, position.amount().neg())
                            .await?
                    }
                    OrderKind::Market => {
                        self.submit_market_order(closing_price, position.amount().neg())
                            .await?
                    }
                    OrderKind::ExchangeMarket => {
                        self.submit_exchange_market_order(closing_price, position.amount().neg())
                            .await?
                    }
                    _ => {
                        unimplemented!()
                    }
                };

                self.tracked_positions.insert(position.id(), active_order);
            }
        }

        Ok(())
    }

    pub async fn submit_limit_order(
        &self,
        closing_price: f64,
        amount: f64,
    ) -> Result<ActiveOrder, BoxError> {
        info!(
            "Submitting exchange limit order of {} {}...",
            amount,
            self.pair.base()
        );
        let order_form = OrderForm::new(&self.pair, closing_price, amount, OrderKind::Limit);

        Ok(self.client.submit_order(order_form).await?)
    }

    pub async fn submit_exchange_limit_order(
        &self,
        closing_price: f64,
        amount: f64,
    ) -> Result<ActiveOrder, BoxError> {
        info!(
            "Submitting exchange limit order of {} {}...",
            amount,
            self.pair.base()
        );
        let order_form =
            OrderForm::new(&self.pair, closing_price, amount, OrderKind::ExchangeLimit);

        Ok(self.client.submit_order(order_form).await?)
    }

    pub async fn submit_market_order(
        &self,
        closing_price: f64,
        amount: f64,
    ) -> Result<ActiveOrder, BoxError> {
        info!(
            "Submitting market order of {} {}...",
            amount,
            self.pair.base()
        );
        let order_form = OrderForm::new(&self.pair, closing_price, amount, OrderKind::Market);

        Ok(self.client.submit_order(order_form).await?)
    }

    pub async fn submit_exchange_market_order(
        &self,
        closing_price: f64,
        amount: f64,
    ) -> Result<ActiveOrder, BoxError> {
        info!(
            "Submitting market order of {} {}...",
            amount,
            self.pair.base()
        );
        let order_form =
            OrderForm::new(&self.pair, closing_price, amount, OrderKind::ExchangeMarket);

        Ok(self.client.submit_order(order_form).await?)
    }

    pub fn update(&self) -> Result<OptionUpdate, BoxError> {
        // TODO: implement me
        Ok((None, None))
    }

    pub fn best_closing_price(&self, position: &Position, order_book: &OrderBook) -> f64 {
        let ask = order_book.lowest_ask();
        let bid = order_book.highest_bid();
        let avg = (bid + ask) / 2.0;
        let delta = (ask - bid) / 10.0;

        let closing_price = {
            let closing_price = {
                if position.is_short() {
                    bid - delta
                } else {
                    ask + delta
                }
            };

            if avg > 9999.0 {
                if position.is_short() {
                    closing_price.ceil()
                } else {
                    closing_price.floor()
                }
            } else {
                closing_price
            }
        };

        closing_price
    }
}

pub struct PairManager {
    pair: SymbolPair,
    price_manager: PriceManagerHandle,
    order_manager: OrderManagerHandle,
    position_manager: PositionManagerHandle,
}

impl PairManager {
    pub fn new(pair: SymbolPair, client: Client) -> Self {
        Self {
            pair: pair.clone(),
            price_manager: PriceManagerHandle::new(pair.clone(), client.clone()),
            order_manager: OrderManagerHandle::new(
                pair.clone(),
                client.clone(),
                Box::new(FastOrderStrategy::default()),
            ),
            position_manager: PositionManagerHandle::new(
                pair.clone(),
                client.clone(),
                Box::new(TrailingStop::new()),
            ),
        }
    }

    pub async fn update_managers(&mut self, tick: u64) -> Result<(), BoxError> {
        let (opt_price_events, opt_price_messages) = self.price_manager.update(tick).await?;
        let (opt_pos_events, opt_pos_messages) = self.position_manager.update(tick).await?;
        let (opt_order_events, opt_order_messages) = self.order_manager.update(tick).await?;

        // TODO: to move into Handler?

        if let Some(messages) = opt_pos_messages {
            for m in messages {
                match m {
                    Message::ClosePosition {
                        position,
                        order_kind,
                    } => {
                        self.order_manager
                            .close_position(position, order_kind)
                            .await?;
                    }
                    _ => {}
                }
            }
        }

        Ok(())
    }
}

pub struct ExchangeManager {
    kind: ExchangeKind,
    pair_managers: Vec<PairManager>,
    client: Client,
}

impl ExchangeManager {
    pub fn new(kind: &ExchangeKind, pairs: &Vec<SymbolPair>) -> Self {
        let client = Client::new(kind);
        let pair_managers = pairs
            .into_iter()
            .map(|x| PairManager::new(x.clone(), client.clone()))
            .collect();

        Self {
            kind: kind.clone(),
            pair_managers,
            client,
        }
    }

    pub async fn update_managers(&mut self, tick: u64) -> Result<(), BoxError> {
        let mut futures: FuturesUnordered<_> = self
            .pair_managers
            .iter_mut()
            .map(|x| x.update_managers(tick))
            .collect();

        // execute the futures
        while let Some(_) = futures.next().await {}

        Ok(())
    }

    // async fn update_position_managers(&mut self, tick: u64) -> Result<OptionUpdate, BoxError> {
    //     let mut res_events = Vec::new();
    //     let mut res_messages = Vec::new();
    //
    //     let mut futures: FuturesUnordered<_> = self
    //         .pair_managers
    //         .iter_mut()
    //         .map(|x| x.position_manager.update(tick))
    //         .collect();
    //
    //     while let Some(future_result) = futures.next().await {
    //         let (opt_events, opt_messages) = future_result?;
    //
    //         if let Some(events) = opt_events {
    //             res_events.extend(events);
    //         }
    //
    //         if let Some(messages) = opt_messages {
    //             res_messages.extend(messages);
    //         }
    //     }
    //
    //     Ok((
    //         (!res_events.is_empty()).then_some(res_events),
    //         (!res_messages.is_empty()).then_some(res_messages),
    //     ))
    // }
    //
    // async fn update_price_managers(&mut self, tick: u64) -> Result<OptionUpdate, BoxError> {
    //     let mut futures: FuturesUnordered<_> = self
    //         .pair_managers
    //         .iter_mut()
    //         .map(|x| x.price_manager.update(tick))
    //         .collect();
    //
    //     while let Some(x) = futures.next().await {}
    //
    //     Ok((None, None))
    // }
}
